#include <cstdlib>
#include <cassert>
#include "queue.h"


DoublyListNode *newDoublyListNode(void *data) {
    DoublyListNode *node = (DoublyListNode *) malloc(sizeof(DoublyListNode));
    node->val = data;
    node->next = NULL;
    node->prev = NULL;
    return node;
}

/* 析构函数 */
void delDoublyListNode(DoublyListNode *node) {
    free(node);
}


/* 构造函数 */
LinkedListDeque *newLinkedListDeque() {
    LinkedListDeque *deque = (LinkedListDeque *) malloc(sizeof(LinkedListDeque));
    deque->front = NULL;
    deque->rear = NULL;
    deque->queSize = 0;
    return deque;
}

/* 析构函数 */
void delLinkedListdeque(LinkedListDeque *deque) {
    // 释放所有节点
    for (int i = 0; i < deque->queSize && deque->front != NULL; i++) {
        DoublyListNode *tmp = deque->front;
        deque->front = deque->front->next;
        free(tmp);
    }
    // 释放 deque 结构体
    free(deque);
}

/* 获取队列的长度 */
int queue_size(LinkedListDeque *deque) {
    return deque->queSize;
}

/* 判断队列是否为空 */
bool queue_is_empty(LinkedListDeque *deque) {
    return (queue_size(deque) == 0);
}

/* 入队 */
void push(LinkedListDeque *deque, void *data, bool isFront) {
    DoublyListNode *node = newDoublyListNode(data);
    // 若链表为空，则令 front 和 rear 都指向node
    if (queue_is_empty(deque)) {
        deque->front = deque->rear = node;
    }
        // 队首入队操作
    else if (isFront) {
        // 将 node 添加至链表头部
        deque->front->prev = node;
        node->next = deque->front;
        deque->front = node; // 更新头节点
    }
        // 队尾入队操作
    else {
        // 将 node 添加至链表尾部
        deque->rear->next = node;
        node->prev = deque->rear;
        deque->rear = node;
    }
    deque->queSize++; // 更新队列长度
}

/* 队首入队 */
void pushFirst(LinkedListDeque *deque, void *data) {
    push(deque, data, true);
}

/* 队尾入队 */
void pushLast(LinkedListDeque *deque, void *data) {
    push(deque, data, false);
}

/* 访问队首元素 */
void *peekFirst(LinkedListDeque *deque) {
    assert(queue_size(deque) && deque->front);
    return deque->front->val;
}

/* 访问队尾元素 */
void *peekLast(LinkedListDeque *deque) {
    assert(queue_size(deque) && deque->rear);
    return deque->rear->val;
}

/* 出队 */
void *pop(LinkedListDeque *deque, bool isFront) {
    if (queue_is_empty(deque))
        return nullptr;
    void *val;
    // 队首出队操作
    if (isFront) {
        val = peekFirst(deque); // 暂存头节点值
        DoublyListNode *fNext = deque->front->next;
        if (fNext) {
            fNext->prev = NULL;
            deque->front->next = NULL;
        }
        delDoublyListNode(deque->front);
        deque->front = fNext; // 更新头节点
    }
        // 队尾出队操作
    else {
        val = peekLast(deque); // 暂存尾节点值
        DoublyListNode *rPrev = deque->rear->prev;
        if (rPrev) {
            rPrev->next = NULL;
            deque->rear->prev = NULL;
        }
        delDoublyListNode(deque->rear);
        deque->rear = rPrev; // 更新尾节点
    }
    deque->queSize--; // 更新队列长度
    return val;
}

/* 队首出队 */
void *popFirst(LinkedListDeque *deque) {
    return pop(deque, true);
}

/* 队尾出队 */
void *popLast(LinkedListDeque *deque) {
    return pop(deque, false);
}